use core::panic;

use crate::{executor::Executor, lexer::{Token, TokenType}};

#[derive(Debug, Clone, PartialEq)]
pub enum ASTPart {
    String(AstString),
    Number(AstNumber),
    Boolean(AstBool),
    Null(AstNull),
    Assigment(AstAssigment),
    Operation(AstOperation),
    VarRead(AstVarRead),
    Call(AstCall),
    VarUpdate(AstVarUpdate),
    Function(AstFunction),
    If(AstIf),
    ElseIf(AstElseIf),
    Else(AstElse),
    While(AstWhile),
    Break(AstBreak),
    For(AstFor),
    Continue(AstContinue),
    Return(AstReturn),
    RustFunction(AstRustFunction),
    NOOP
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstString {
    pub value: String,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstNumber {
    pub value: i64,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstBool {
    pub value: bool,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstNull {
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstAssigment {
    pub variable: String,
    pub value: Box<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstOperation {
    pub operator: String,
    pub left: Box<ASTPart>,
    pub right: Box<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstVarRead {
    pub variable: String,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstCall {
    pub function: Box<ASTPart>,
    pub args: Vec<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstVarUpdate {
    pub variable: String,
    pub value: Box<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstFunction {
    pub args: Vec<String>,
    pub body: Vec<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstIf {
    pub condition: Box<ASTPart>,
    pub body: Vec<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstElseIf {
    pub condition: Box<ASTPart>,
    pub body: Vec<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstElse {
    pub body: Vec<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstWhile {
    pub condition: Box<ASTPart>,
    pub body: Vec<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstBreak {
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstFor {
    pub init: Box<ASTPart>,
    pub condition: Box<ASTPart>,
    pub update: Box<ASTPart>,
    pub body: Vec<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstContinue {
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstReturn {
    pub value: Box<ASTPart>,
    pub pos: usize
}
#[derive(Debug, Clone, PartialEq)]
pub struct AstRustFunction {
    pub func: fn(&mut Executor, Vec<ASTPart>) -> ASTPart,
    pub pos: usize
}

fn is_end(input: &Token, end: &Vec<Token>) -> bool {
    for token in end {
        if input.typ == token.typ && (token.value == "" || input.value == token.value) {
            return true;
        }
    }
    return false;
}

fn read_call(variable: ASTPart, pos: &mut usize, input: &Vec<Token>) -> ASTPart {
    let mut args: Vec<ASTPart> = vec![];
    *pos += 1;
    let start_pos = input[*pos-1].pos;
    while pos < &mut input.len() {
        let token = &input[*pos];
        if token.typ == TokenType::SEPARATOR && token.value == "," {
            *pos += 1;
            continue;            
        }
        if token.typ == TokenType::SEPARATOR && token.value == String::from(")") {
            *pos += 1;
            break;
        }
        let ends: Vec<Token> = vec![
            Token { typ: TokenType::SEPARATOR, value: String::from(","), pos: 0 },
            Token { typ: TokenType::SEPARATOR, value: String::from(")"), pos: 0 }
        ];
        let arg = read_exp(pos, input, &ends, &ends);
        args.push(arg);
    }
    return ASTPart::Call(AstCall { function: Box::new(variable), args: args, pos: start_pos });
}

fn operator_precedence(op: &str) -> i64 {
    match op {
        "|" | "&" => 1,
        "==" | "!=" | "<=" | ">=" => 2,
        "+" | "-" => 3,
        "*" | "/" | "%" => 4,
        "^" => 5,
        _ => 0
    }
}

fn shunt(input: Vec<ASTPart>) -> ASTPart {
    let mut output: Vec<ASTPart> = vec![];
    let mut stack: Vec<ASTPart> = vec![];
    for part in input {
        match &part {
            ASTPart::String(_) => {
                output.push(part);
            },
            ASTPart::Number(_) => {
                output.push(part);
            },
            ASTPart::Call(_) => {
                stack.push(part);
            },
            ASTPart::VarRead(_) => {
                output.push(part);
            },
            ASTPart::Boolean(_) => {
                output.push(part);
            },
            ASTPart::Null(_) => {
                output.push(part);
            },
            ASTPart::Function(_) => {
                output.push(part);
            },
            ASTPart::Operation(op) => {
                if *op.left != ASTPart::NOOP && *op.right != ASTPart::NOOP {
                    output.push(part);
                    continue;
                }
                while stack.len() > 0 {
                    let top = &stack[stack.len()-1];
                    match top {
                        ASTPart::Operation(top_op) => {
                            if operator_precedence(&top_op.operator) >= operator_precedence(&op.operator) {
                                output.push(stack.pop().unwrap());
                            } else {
                                break;
                            }
                        },
                        _ => break
                    }
                }
                stack.push(part);
            },
            _ => {}
        }
    }
    while stack.len() > 0 {
        output.push(stack.pop().unwrap());
    }
    let mut i = 0;
    while i < output.len() {
        match &output[i] {
            ASTPart::Operation(op) => {
                if *op.left != ASTPart::NOOP && *op.right != ASTPart::NOOP {
                    i += 1;
                    continue;
                }
                if op.operator == "!" {
                    println!("{:?}", output);
                    if i < 1 {
                        panic!("Unexpected operation at {}", op.pos);
                    }
                    let left = output[i-1].clone();
                    output[i] = ASTPart::Operation(AstOperation {
                        operator: op.operator.clone(),
                        left: Box::new(left),
                        right: Box::new(ASTPart::NOOP),
                        pos: op.pos,
                    });
                    output.remove(i-1);
                } else {
                    if i < 2 {
                        panic!("Unexpected operation at {}", op.pos);
                    }
                    let left = output[i-2].clone();
                    let right = output[i-1].clone();
                    output[i] = ASTPart::Operation(AstOperation {
                        operator: op.operator.clone(),
                        left: Box::new(left),
                        right: Box::new(right),
                        pos: op.pos,
                    });
                    output.remove(i-2);
                    output.remove(i-2);
                    i -= 1;
                }
            }
            _ => {
                i += 1;
            }
        }
    }
    return output[0].clone();
}
fn read_function(input: &Vec<Token>, pos: &mut usize, with_args: bool) -> ASTPart {
    let mut args: Vec<String> = vec![];
    let start_pos = input[*pos].pos;
    if with_args {
        if input[*pos].typ != TokenType::SEPARATOR || input[*pos].value != "(" {
            panic!("Expected ( at {}", input[*pos].pos);
        }
        *pos += 1;
        while pos < &mut input.len() {
            let token = &input[*pos];
            if token.typ == TokenType::SEPARATOR && token.value == ")" {
                break;
            }
            *pos += 1;
            if token.typ == TokenType::SEPARATOR && token.value == "," {
                continue;
            }
            if token.typ == TokenType::IDENTIFIER {
                args.push(token.value.clone());
            } else {
                panic!("Unexpected {:?}({}) at {}", token.typ, token.value, token.pos);
            }
        }
        if input[*pos].typ != TokenType::SEPARATOR || input[*pos].value != ")" {
            panic!("Unexpected end of arguments at {}", input[*pos].pos);
        }
        *pos += 1;
    }
    if input[*pos].typ != TokenType::SEPARATOR || input[*pos].value != "{" {
        panic!("Expected {{ at {}", input[*pos].pos);
    }
    *pos += 1;
    let op_ends: Vec<Token> = vec![
        Token { typ: TokenType::OPEND, value: String::from("\n"), pos: 0 },
        Token { typ: TokenType::OPEND, value: String::from(";"), pos: 0 },
        Token { typ: TokenType::SEPARATOR, value: String::from("}"), pos: 0 }
    ];
    let parse_ends: Vec<Token> = vec![
        Token { typ: TokenType::SEPARATOR, value: String::from("}"), pos: 0 }
    ];
    let body = parse_internal(input, &op_ends, &parse_ends, pos);
    if input[*pos].typ != TokenType::SEPARATOR || input[*pos].value != "}" {
        panic!("Unexpected end of function at {}", input[*pos].pos);
    }
    *pos += 1;
    return ASTPart::Function(AstFunction { args: args, body: body, pos: start_pos });
}
fn read_exp(pos: &mut usize, input: &Vec<Token>, ends: &Vec<Token>, parse_ends: &Vec<Token>) -> ASTPart {
    let mut expressions: Vec<ASTPart> = vec![];
    while pos < &mut input.len() {
        let token = &input[*pos];
        let mut next_token = &Token {
            typ: TokenType::OPEND,
            value: String::from("END"),
            pos: 0
        };
        if *pos+1 < input.len() {
            next_token = &input[*pos+1]
        }
        if is_end(token, &parse_ends) {
            break;
        }
        *pos += 1;
        if is_end(token, ends) {
            break;
        }
        if token.typ == TokenType::STRING {
            expressions.push(ASTPart::String(AstString { value: token.value.clone(), pos: token.pos }));
        } else if token.typ == TokenType::NUMBER {
            expressions.push(ASTPart::Number(AstNumber { value: token.value.parse().unwrap(), pos: token.pos }));
        } else if token.typ == TokenType::KEYWORD {
            if token.value == "piszv" {
                expressions.push(ASTPart::Boolean(AstBool { value: true, pos: token.pos }));
            } else if token.value == "nem piszv" {
                expressions.push(ASTPart::Boolean(AstBool { value: false, pos: token.pos }));
            } else if token.value == "nincs hám" {
                expressions.push(ASTPart::Null(AstNull { pos: token.pos }));
            } else if token.value == "lőcsve" {
                let func = read_function(input, pos, true);
                expressions.push(func);
            } else {
                panic!("Unexpected {:?}({}) at {}", token.typ, token.value, token.pos);
            }
        } else if token.typ == TokenType::IDENTIFIER {
            if next_token.typ == TokenType::SEPARATOR && next_token.value == "(" {
                let var = ASTPart::VarRead(AstVarRead { variable: token.value.clone(), pos: token.pos });
                expressions.push(read_call(var, pos, input));
            } else {
                expressions.push(ASTPart::VarRead(AstVarRead { variable: token.value.clone(), pos: token.pos }));
            }
        } else if token.typ == TokenType::OPERATOR {
            expressions.push(ASTPart::Operation(AstOperation { operator: token.value.clone(), left: Box::new(ASTPart::NOOP), right: Box::new(ASTPart::NOOP), pos: token.pos }));
        } else if token.typ == TokenType::SEPARATOR {
            //We check for () and then send the into read_exp again, so we recursively parse the expression
            if token.value == "(" {
                let ends: Vec<Token> = vec![
                    Token { typ: TokenType::SEPARATOR, value: String::from(")"), pos: 0 }
                ];
                let exp = read_exp(pos, input, &ends, &ends);
                if input[*pos].typ == TokenType::SEPARATOR && input[*pos].value == ")" {
                    *pos += 1;
                } else {
                    panic!("Unclosed parenthesis at {}", token.pos);
                }
                expressions.push(exp);
            } else {
                panic!("Unexpected {:?}({}) at {}", token.typ, token.value, token.pos);
            }
        } else {
            panic!("Unexpected {:?}({}) at {}", token.typ, token.value, token.pos);
        }
    }
    let shunted = shunt(expressions);
    return shunted;
}

fn next_operation(pos: &mut usize, input: &Vec<Token>, op_ends: &Vec<Token>, parse_ends: &Vec<Token>) -> ASTPart {
    let token = &input[*pos];
    let mut next_token = &Token {
        typ: TokenType::OPEND,
        value: String::from("END"),
        pos: 0
    };
    if *pos+1 < input.len() {
        next_token = &input[*pos+1]
    }
    if is_end(token, &parse_ends) {
        return ASTPart::NOOP;
    }
    *pos += 1;
    if is_end(token, &op_ends) {
        return ASTPart::NOOP;
    }
    if token.typ == TokenType::KEYWORD {
        if token.value == "gethelj" {
            let variable = &input[*pos];
            *pos += 1;
            if variable.typ != TokenType::IDENTIFIER {
                panic!("Unexpected {:?} at {}", variable.typ, variable.pos)
            }
            let eq = &input[*pos];
            if eq.typ == TokenType::SEPARATOR && eq.value == "=" {
                *pos += 1;
            }
            let value = read_exp(pos, input, op_ends, parse_ends);
            return ASTPart::Assigment(AstAssigment { variable: variable.value.clone(), value: Box::new(value), pos: token.pos });
        } else if token.value == "ha geny" {
            if next_token.typ != TokenType::SEPARATOR || next_token.value != "(" {
                panic!("Expected ( at {}", token.pos);
            }
            *pos += 1;
            let condition_end = vec![
                Token { typ: TokenType::SEPARATOR, value: String::from(")"), pos: 0 }
            ];
            let condition = read_exp(pos, input, &condition_end, &condition_end);
            if input[*pos].typ != TokenType::SEPARATOR || input[*pos].value != ")" {
                panic!("Unexpected end of condition at {}", token.pos);
            }
            *pos += 1;
            let body = read_function(input, pos, false);
            let real_body = match body {
                ASTPart::Function(func) => func.body,
                _ => panic!("Expected function body at {}", token.pos)
            };
            return ASTPart::If(AstIf { condition: Box::new(condition), body: real_body, pos: token.pos });
        } else if token.value == "amíg geny" {
            if next_token.typ != TokenType::SEPARATOR || next_token.value != "(" {
                panic!("Expected ( at {}", token.pos);
            }
            *pos += 1;
            let condition_end = vec![
                Token { typ: TokenType::SEPARATOR, value: String::from(")"), pos: 0 }
            ];
            let condition = read_exp(pos, input, &condition_end, &condition_end);
            if input[*pos].typ != TokenType::SEPARATOR || input[*pos].value != ")" {
                panic!("Unexpected end of condition at {}", token.pos);
            }
            *pos += 1;
            let body = read_function(input, pos, false);
            let real_body = match body {
                ASTPart::Function(func) => func.body,
                _ => panic!("Expected function body at {}", token.pos)
            };
            return ASTPart::While(AstWhile { condition: Box::new(condition), body: real_body, pos: token.pos });
        } else if token.value == "kraf" {
            return ASTPart::Break(AstBreak { pos: token.pos });
        } else if token.value == "kopva" {
            if next_token.typ != TokenType::SEPARATOR || next_token.value != "(" {
                panic!("Expected ( at {}", token.pos);
            }
            *pos += 1;
            let ends = vec![
                Token { typ: TokenType::SEPARATOR, value: String::from(")"), pos: 0 },
                Token { typ: TokenType::OPEND, value: String::from(";"), pos: 0 }
            ];
            let init = parse_internal(input, &ends, &ends, pos);
            if init.len() != 1 {
                panic!("Only one expression is expected for init at {}", token.pos);
            }
            if input[*pos].typ != TokenType::OPEND || input[*pos].value != ";" {
                panic!("Unexpected end of init at {}", token.pos);
            }
            *pos += 1;
            let condition = read_exp(pos, input, &ends, &ends);
            if input[*pos].typ != TokenType::OPEND || input[*pos].value != ";" {
                panic!("Unexpected end of condition at {}", token.pos);
            }
            *pos += 1;
            let update = parse_internal(input, &ends, &ends, pos);
            if update.len() != 1 {
                panic!("Only one expression is expected for update at {}", token.pos);
            }
            if input[*pos].typ != TokenType::SEPARATOR || input[*pos].value != ")" {
                panic!("Unexpected end of update at {}", token.pos);
            }
            *pos += 1;
            let body = read_function(input, pos, false);
            let real_body = match body {
                ASTPart::Function(func) => func.body,
                _ => panic!("Expected function body at {}", token.pos)
            };
            return ASTPart::For(AstFor { init: Box::new(init[0].clone()), condition: Box::new(condition), update: Box::new(update[0].clone()), body: real_body, pos: token.pos });
        } else if token.value == "szard le" {
            return ASTPart::Continue(AstContinue { pos: token.pos });
        } else if token.value == "ha nem geny akkor geny" {
            if next_token.typ != TokenType::SEPARATOR || next_token.value != "(" {
                panic!("Expected ( at {}", token.pos);
            }
            *pos += 1;
            let condition_end = vec![
                Token { typ: TokenType::SEPARATOR, value: String::from(")"), pos: 0 }
            ];
            let condition = read_exp(pos, input, &condition_end, &condition_end);
            if input[*pos].typ != TokenType::SEPARATOR || input[*pos].value != ")" {
                panic!("Unexpected end of condition at {}", token.pos);
            }
            *pos += 1;
            let body = read_function(input, pos, false);
            let real_body = match body {
                ASTPart::Function(func) => func.body,
                _ => panic!("Expected function body at {}", token.pos)
            };
            return ASTPart::ElseIf(AstElseIf { condition: Box::new(condition), body: real_body, pos: token.pos });
        } else if token.value == "ha nem geny" {
            let body = read_function(input, pos, false);
            let real_body = match body {
                ASTPart::Function(func) => func.body,
                _ => panic!("Expected function body at {}", token.pos)
            };
            return ASTPart::Else(AstElse { body: real_body, pos: token.pos });
        } else if token.value == "reti" {
            if is_end(next_token, op_ends) || is_end(next_token, parse_ends) {
                return ASTPart::Return(AstReturn { value: Box::new(ASTPart::Null(AstNull { pos: token.pos })), pos: token.pos });
            } else {
                let value = read_exp(pos, input, op_ends, parse_ends);
                return ASTPart::Return(AstReturn { value: Box::new(value), pos: token.pos });
            }
        } else {
            panic!("Unexpected {:?}({}) at {}", token.typ, token.value, token.pos);
        }
    } else if token.typ == TokenType::IDENTIFIER {
        if next_token.typ == TokenType::SEPARATOR && next_token.value == "(" {
            let var = ASTPart::VarRead(AstVarRead { variable: token.value.clone(), pos: token.pos });
            return read_call(var, pos, input);
        } else if next_token.typ == TokenType::SEPARATOR && next_token.value == "=" {
            *pos += 1;
            let value = read_exp(pos, input, op_ends, parse_ends);
            return ASTPart::VarUpdate(AstVarUpdate { variable: token.value.clone(), value: Box::new(value), pos: token.pos });
        } else {
            panic!("Unexpected {:?}({}) at {}", token.typ, token.value, token.pos);
        }
    } else {
        panic!("Unexpected {:?}({}) at {}", token.typ, token.value, token.pos);
    }
}

fn parse_internal(input: &Vec<Token>, op_ends: &Vec<Token>, parse_ends: &Vec<Token>, pos: &mut usize) -> Vec<ASTPart> {
    let mut out: Vec<ASTPart> = vec![];
    while *pos < input.len() {
        let op = next_operation(pos, &input, &op_ends, &parse_ends);
        match op {
            ASTPart::NOOP => {},
            _ => {
                out.push(op);
            }
        }
        if is_end(&input[*pos], &parse_ends) {
            break;
        }
    }
    return out;
}
pub fn parse(input: Vec<Token>) -> Vec<ASTPart> {
    let op_ends: Vec<Token> = vec![
        Token { typ: TokenType::OPEND, value: String::from("\n"), pos: 0 },
        Token { typ: TokenType::OPEND, value: String::from(";"), pos: 0 },
        Token { typ: TokenType::OPEND, value: String::from("EOF"), pos: 0 }
    ];
    let parse_ends: Vec<Token> = vec![
        Token { typ: TokenType::OPEND, value: String::from("EOF"), pos: 0 }
    ];
    let out = parse_internal(&input, &op_ends, &parse_ends, &mut 0);
    return out;
}